"""
Calculate shear from a Nortek AD2CP mounted on a glider.
Compatible with any glider as long as the ADCP is returning data for the 4 beams.
TODO: make compatible with the 3-beam configuration.


gliderad2cp.process_shear
--------------------------
process
    The main function which converts beam velocity data shear in the ENU directions.
load_data
    Loads AD2CP netcdf file and creates supplemental variables based on input glider data.
_velocity_soundspeed_correction
    Correct for salinity error in the soundspeed calculation.
_quality_control_velocities
    Removes bad velocity measurements based on velocity, amplitude and correlation criteria defined in options.
_determine_velocity_measurement_depths
    Determines what depth each velocity measurement was taken at for each beam, account for glider attitude.
_regrid_beam_velocities_to_isobars
    Regrids beam velocities onto isobars to avoid shear smearing in the final shear data.
_rotate_BEAMS_to_XYZ
    Transforms velocity measurements from along-beam coordinates to XYZ coordinates.
_rotate_XYZ_to_ENU
    Rotates velocity measurements in XYZ coordinates into ENU coordinates.
    

Notes
-------
.process() runs the following functions in this order

1. load_data
2. (correct_heading - Awaiting publication, contact Bastien Queste for access)
3. _velocity_soundspeed_correction
4. _quality_control_velocities
5. _determine_velocity_measurement_depths 
6. _regrid_beam_velocities_to_isobars
7. _rotate_BEAMS_to_XYZ
8. _rotate_XYZ_to_ENU
    
"""

import warnings
from glob import glob
import pandas as pd
import numpy as np
import datetime
import xarray as xr
import gsw
from scipy.interpolate import interp1d
from .tools import plog, interp, get_options

warnings.filterwarnings(action='ignore', message='Mean of empty slice')
warnings.filterwarnings(action='ignore', message='invalid value encountered in divide')
warnings.filterwarnings(action='ignore', message='invalid value encountered in true_divide')
warnings.filterwarnings(action='ignore', message='Degrees of freedom <= 0 for slice.')


"""
Data loading functions
"""
def load_data(adcp_file_path, glider_file_path, options):
    """
    Loads AD2CP netcdf file and creates supplemental variables based on input glider data.


    Inputs
    ----------
    adcp_file_path : str
        Path to the AD2CP netcdf files created by the Nortek MIDAS software. Can handle wildcards through glob.
    glider_file_path : str
        Path to the glider file containing time, latitude, longitude, soundspeed, and profile_num.
        Can be csv, pandas or xarray. Can also pass pandas or xarray dataframes directly.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function.


    Outputs
    -------
    ADCP : xr.Dataset
        xr.Dataset of the AD2CP netcdf with necessary glider variables interpolated onto the time dimension, with some minor cleaning of coordinates.
    options : dict
        Same as the input options with the very important addition of ADCP mounting direction if the 'auto' option is set for 'ADCP_mounting_direction'.
    glider_data : pd.Dataframe
        pd.Dataframe of the input glider data. Serves no real purpose but may be helpful in debugging. Discarded when called from .process()
    
    """
    def __load_glider_data(glider_file):
        """
        Loads glider data.


        Inputs
        ----------
        glider_file_path : str
            Path to the glider file containing time, latitude, longitude, soundspeed, and profile_num. Can be csv, pandas or xarray.


        Outputs
        -------
        glider_data : pd.Dataframe
            pd.Dataframe of the input glider data. Serves no real purpose but may be helpful in debugging. Discarded when called from .process()

        """
        if type(glider_file) is pd.core.frame.DataFrame:
            data = glider_file
        elif type(glider_file) is xr.core.dataset.Dataset:
            data = glider_file.to_dataframe()
            if 'time' not in data:
                data['time'] = data.index        
            data['date_float'] = data['time'].values.astype('float')
            if 'profile_index' in data:
                data['profile_number'] = data['profile_index']
            if 'profile_num' in data:
                data['profile_number'] = data['profile_num']
        elif str(glider_file)[-4:] == '.csv':
            data = pd.read_table(glider_file, parse_dates=['time'])
        elif str(glider_file)[-4:] == '.pqt':
            data = pd.read_parquet(glider_file)
        elif str(glider_file)[-3:] == '.nc':
            data = xr.open_dataset(glider_file).to_dataframe()
            if 'time' not in data:
                data['time'] = data.index        
            data['date_float'] = data['time'].values.astype('float')
            if 'profile_index' in data:
                data['profile_number'] = data['profile_index']
            if 'profile_num' in data:
                data['profile_number'] = data['profile_num']
        else:
            raise(ValueError(f"could not parse input dataset {glider_file}"))
        sel_cols = [
            'time',
            'temperature',
            'salinity',
            'latitude',
            'longitude',
            'profile_number',
            'pressure',
            'dive_num'
        ]
        valid_cols = list(set(list(data)).intersection(sel_cols))
        data = data[valid_cols]
        time_ms = data.time.values
        if time_ms.dtype != '<M8[ns]':
            divisor = 1e3
            if time_ms.dtype == '<M8[us]':
                divisor = 1e6
            time_float = time_ms.astype('float') / divisor
            base = datetime.datetime(1970, 1, 1)
            time_ms = []
            for seconds in time_float:
                time_ms.append(base + datetime.timedelta(seconds=seconds + 0.0000001))
            time_nanoseconds = pd.to_datetime(time_ms)
            data['time'] = time_nanoseconds
            
            
        data['date_float'] = data['time'].values.astype('float')
        p = data['pressure']
        SA = gsw.conversions.SA_from_SP(data.salinity, p, data.longitude, data.latitude)
        CT = gsw.CT_from_t(SA, data['temperature'], p)
        data['soundspeed'] = gsw.sound_speed(SA, CT, p)
        data.index = data.time
        data.index.name = None
        return data
    
    ## Load glider data
    plog('Loading glider data.')
    glider_data = __load_glider_data(glider_file_path)
    
    ## Load ADCP netcdf file(s)
    plog('Loading ADCP data.')
    ADCP = xr.open_mfdataset(adcp_file_path, group='Data/Average')
    ADCP_settings = xr.open_mfdataset(glob(adcp_file_path)[0], group='Config')
    ADCP.attrs = ADCP_settings.attrs
    
    plog('Merging glider data into ADCP dataset.')
    adcp_time_float = ADCP.time.values.astype('float')
    glider_time_float = glider_data['date_float'].values
    ADCP = ADCP.drop_vars(['MatlabTimeStamp']) # In protest of closed source software.
    
    # Coordinates
    ADCP = ADCP.assign_coords(Latitude  = ('time', interp(glider_time_float, glider_data['latitude'], adcp_time_float)))
    ADCP = ADCP.assign_coords(Longitude = ('time',interp(glider_time_float, glider_data['longitude'], adcp_time_float)))
    
    ADCP['Latitude'].attrs = {"units": 'Decimal degrees', 'description': f'Latitude from glider data.'}
    ADCP['Longitude'].attrs = {"units": 'Decimal degrees', 'description': f'Longitude from glider data.'}

    # Profile and depth
    ADCP = ADCP.assign_coords(profile_number=('time', np.round(interp(glider_time_float, glider_data['profile_number'], adcp_time_float))))
    ADCP = ADCP.assign_coords(Depth=('time', -gsw.z_from_p(ADCP['Pressure'].values, ADCP['Latitude'].values)) )
    
    ADCP['glider_soundspeed'] = ('time', interp(glider_time_float, glider_data['soundspeed'], adcp_time_float) , {"units": 'm.s-1', 'description': f'Sound velocity in water from glider data.'})
    ADCP['glider_salinity'] = ('time', interp(glider_time_float, glider_data['salinity'], adcp_time_float) , {"units": 'PSU', 'description': f'Practical salinity from glider data.'})
    ADCP['glider_temperature'] = ('time', interp(glider_time_float, glider_data['temperature'], adcp_time_float) , {"units": 'degrees C', 'description': f'ITS-90 temperature from glider data.'})

    # Get rid of pointless dimensions and make them coordinates instead
    ADCP = ADCP.assign_coords( bin=('Velocity Range', np.arange(len(ADCP['Velocity Range'].values)) ) )
    ADCP = ADCP.swap_dims({'Velocity Range': 'bin'})

    ADCP = ADCP.assign_coords( bin=('Correlation Range', np.arange(len(ADCP['Correlation Range'].values)) ) )
    ADCP = ADCP.swap_dims({'Correlation Range': 'bin'})

    ADCP = ADCP.assign_coords( bin=('Amplitude Range', np.arange(len(ADCP['Amplitude Range'].values)) ) )
    ADCP = ADCP.swap_dims({'Amplitude Range': 'bin'})

    if options['ADCP_mounting_direction'] == 'auto':
        # Detect if ADCP is top mounted using the accelerometer
        if np.nanmedian(ADCP.AccelerometerZ.values) < 0:
            options['ADCP_mounting_direction'] = 'bottom'
        else:
            options['ADCP_mounting_direction'] = 'top'
        plog(f'ADCP is '+options['ADCP_mounting_direction']+'-mounted according to the accelerometer.')

    return ADCP, options, glider_data


"""
QC functions
"""
def _velocity_soundspeed_correction(ADCP):
    """
    Corrects along beam velocity data for lack of salinity measurements in ADCP default soundspeed settings.


    Inputs
    ----------
    ADCP : xr.Dataset
        GliderAD2CP format xr.Dataset produced by the gliderad2cp.load_data() function.


    Outputs
    -------
    ADCP : xr.Dataset
        Same as input, with correction for sound speed applied to along beam velocities.
    
    """
    if 'NoSal_SpeedOfSound' not in ADCP:
        plog('Performing sound speed correction.')
        ADCP = ADCP.rename({'SpeedOfSound': 'NoSal_SpeedOfSound'})
        ADCP = ADCP.rename({'glider_soundspeed': 'SpeedOfSound'})
        for beam in ['1', '2', '3', '4']:
            # V_new = V_old * (c_new/c_old)
            ADCP['VelocityBeam' + beam] = ADCP['VelocityBeam' + beam] * (
                ADCP['SpeedOfSound'] / ADCP['NoSal_SpeedOfSound']
            )
            plog('    Corrected beam ' + beam + ' velocity for sound speed.')
    else:
        plog('Speed of sound correction already applied')
    return ADCP


def _quality_control_velocities(ADCP, options):
    """
    Removes bad velocity measurements based on velocity, amplitude and correlation criteria defined in options.


    Inputs
    ----------
    ADCP : xr.Dataset
        GliderAD2CP format xr.Dataset produced by the gliderad2cp.load_data() function.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function. Requires 'ADCP_mounting_direction' to be either 'top' or 'bottom'.


    Outputs
    -------
    ADCP : xr.Dataset
        Same as input with bad velocity data replaced with NaNs.
    
    """
    prct = lambda x : '%.2f'%(np.count_nonzero(x) * 100 / n)
    plog('Removing data exceeding correlation, amplitude and velocity thresholds.')
    
    noise_floor = np.nanpercentile(np.concatenate([
        ADCP.AmplitudeBeam1.values,
        ADCP.AmplitudeBeam2.values,
        ADCP.AmplitudeBeam3.values,
        ADCP.AmplitudeBeam4.values
        ]).flatten(),[0.5])
    
    for beam in ['1', '2', '3', '4']:
        
        # Calculate correlation mask
        C = ADCP['CorrelationBeam' + beam].values.copy()
        n = len(C.flatten()) # For percentage calculation above. 
        ind = C < options['QC_correlation_threshold']
        C[ind] = np.nan
        C[np.isfinite(C)] = 1
        plog('    Beam ' + beam + ' correlation: ' + prct(ind) + '% removed')

        # Calculate amplitude mask
        A = ADCP['AmplitudeBeam' + beam].values.copy()
        ind = (A > options['QC_amplitude_threshold']) | (A < noise_floor + options['QC_SNR_threshold'])
        A[ind] = np.nan
        A[np.isfinite(A)] = 1
        plog('    Beam ' + beam + ' amplitude: ' + prct(ind) + '% removed')

        # Calculate velocity mask
        V = ADCP['VelocityBeam' + beam].values.copy()
        ind = np.abs(V) > options['QC_velocity_threshold']
        V[ind] = np.nan
        V[np.isfinite(V)] = 1
        plog('    Beam ' + beam + ' velocity: ' + prct(ind) + '% removed')
        
        # Remove bad data
        ADCP['VelocityBeam' + beam] = ADCP['VelocityBeam' + beam] * C * A * V

    return ADCP


"""
Remapping functions
"""
def _determine_velocity_measurement_depths(ADCP, options):    
    """
    Determines what depth each velocity measurement was taken at for each beam, account for glider attitude.


    Inputs
    ----------
    ADCP : xr.Dataset
        GliderAD2CP format xr.Dataset produced by the gliderad2cp.load_data() function.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function. Requires 'ADCP_mounting_direction' to be either 'top' or 'bottom'.


    Outputs
    -------
    ADCP : xr.Dataset
        Same as input with additional D1, D2, D3 and D4 depth variabiles.
    
    """
    H = ADCP['Heading'] + options['heading_offset']
    P = ADCP['Pitch'] + options['pitch_offset']
    R = ADCP['Roll'] + options['roll_offset']
    
    
    # All *_Range coordinates are distance along beam. Verified with data.
    if options['ADCP_mounting_direction'] == 'top':
        direction = 1
        theta_rad_1 = np.arccos(
            np.cos(np.deg2rad(47.5 - P)) * np.cos(np.deg2rad(R))
        )
        theta_rad_2 = np.arccos(
            np.cos(np.deg2rad(25 - R)) * np.cos(np.deg2rad(P))
        )
        theta_rad_3 = np.arccos(
            np.cos(np.deg2rad(47.5 + P)) * np.cos(np.deg2rad(R))
        )
        theta_rad_4 = np.arccos(
            np.cos(np.deg2rad(25 + R)) * np.cos(np.deg2rad(P))
        )
    else:
        direction = -1
        theta_rad_1 = np.arccos(
            np.cos(np.deg2rad(47.5 + P)) * np.cos(np.deg2rad(R))
        )
        theta_rad_2 = np.arccos(
            np.cos(np.deg2rad(25 + R)) * np.cos(np.deg2rad(P))
        )
        theta_rad_3 = np.arccos(
            np.cos(np.deg2rad(47.5 - P)) * np.cos(np.deg2rad(R))
        )
        theta_rad_4 = np.arccos(
            np.cos(np.deg2rad(25 - R)) * np.cos(np.deg2rad(P))
        )
    
    # The above functions return angles of each beam from the UP direction based on the attitude of the glider.
    # Technically, if the glider is pitched 17.4 degrees, 3 beams are at 30.1 degrees.
    # To remap bins, we need to understand how they are mapped. Is the range equal to distance along beam, or distance along
    # an instrument relative Z-axis.

    # Following pers. comm. with Nortek staff, it has been confirmed that the Velocity Range values are equal to distances
    # from the ADCP along a Z-axis. This means that distance along beam needs to be remapped by the cosine of the beam-to-Z-axis cosine.
    # HOWEVER counterintuitively - they did not use 30.1 degrees. See below correspondence:
    #
    # Sven Nylund<Sven.Nylund@nortekgroup.com> : Sent 27 November 2024 10:28 Swedish time to Bastien Queste.
    # "Sorry for the confusion regarding slant angle yesterday, the correction for slant angle is done at a system level so 1 MHz
    # instruments on the AD2CP platform (like your glider unit) are corrected for a nominal 25-degree slant angle whenever more 
    # than two beams are enabled for measurement. So, in your case, the along beam cell size will be 2.21 m when you set the cell
    # size to 2 m. Like I said yesterday, there is no correction of this based on tilt measurements. There is neither any correction 
    # of the cell size with regards to the sound velocity, we do the time gating based on a nominal sound velocity of 1500 m/s." 
    # 
    # So it is hard-coded for 25 degrees, which doesn't really mesh with the geometry but it's fine as long as we know.
    # It will be important to keep an eye on firmware updates in case they adjust this down the line as the information is not 
    # included in any attributes.
    z_bin_distance = ADCP['Velocity Range'].values/np.cos(np.deg2rad(25))

    ADCP['D1'] = (
        ['time', 'bin'],
        np.tile(ADCP['Depth'], (len(ADCP.bin), 1)).T
        - direction
        * np.tile(z_bin_distance, (len(ADCP.time), 1))
        * np.tile(np.cos(theta_rad_1), (len(ADCP.bin), 1)).T,
        {"units": 'm', 'description': f'Depth of measurement cell along beam 1.'}
    )
    ADCP['D2'] = (
        ['time', 'bin'],
        np.tile(ADCP['Depth'], (len(ADCP.bin), 1)).T
        - direction
        * np.tile(z_bin_distance, (len(ADCP.time), 1))
        * np.tile(np.cos(theta_rad_2), (len(ADCP.bin), 1)).T,
        {"units": 'm', 'description': f'Depth of measurement cell along beam 2.'}
    )
    ADCP['D3'] = (
        ['time', 'bin'],
        np.tile(ADCP['Depth'], (len(ADCP.bin), 1)).T
        - direction
        * np.tile(z_bin_distance, (len(ADCP.time), 1))
        * np.tile(np.cos(theta_rad_3), (len(ADCP.bin), 1)).T,
        {"units": 'm', 'description': f'Depth of measurement cell along beam 3.'}
    )
    ADCP['D4'] = (
        ['time', 'bin'],
        np.tile(ADCP['Depth'], (len(ADCP.bin), 1)).T
        - direction
        * np.tile(z_bin_distance, (len(ADCP.time), 1))
        * np.tile(np.cos(theta_rad_4), (len(ADCP.bin), 1)).T,
        {"units": 'm', 'description': f'Depth of measurement cell along beam 3.'}
    )
    
    return ADCP


def _regrid_beam_velocities_to_isobars(ADCP, options):
    """
    Regrids beam velocities onto isobars to avoid shear smearing in the final shear data.


    Inputs
    ----------
    ADCP : xr.Dataset
        GliderAD2CP format xr.Dataset produced by the gliderad2cp.load_data() function.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function. Requires 'ADCP_mounting_direction' to be either 'top' or 'bottom'.


    Outputs
    -------
    ADCP : xr.Dataset
        Same as input but with the additional gridded_bin dimension (defined as range from the glider), 
        and the V1, V2, V3, V4 velocities which are velocities from beams 1-4 projected onto isobars.
    
    """
    if options['velocity_regridding_distance_from_glider'] == 'auto':
        ## This is to avoid shear smearing because of tilted ADCP
        if options['ADCP_mounting_direction'] == 'top':
            direction = 1
        else:
            direction = -1
        bin_size = ADCP.attrs['avg_cellSize']
        blanking_distance = ADCP.attrs['avg_blankingDistance']
        n_bins = len(ADCP.bin.values)
        max_distance = blanking_distance + n_bins * bin_size + (0.5 * bin_size)
        depth_offsets = np.arange(0, max_distance + bin_size, bin_size / 2) * direction
    else:
        depth_offsets = options['velocity_regridding_distance_from_glider']

    plog('Using the following depth offsets when regridding velocities onto isobars:')
    plog(depth_offsets)
    
    ## Extract to np array for speed
    for beam in ['1', '2', '3', '4']:
        plog(f'    Regridding beam {beam} onto isobars.')

        def __interp1d_np(x, y):
            _gd = np.isfinite(y)
            if np.count_nonzero(_gd) > 1:
                xi = interp1d(x[_gd], y[_gd], bounds_error=False, fill_value=np.nan)(
                    depth_offsets
                )
            else:
                xi = depth_offsets * np.nan
            return xi

        ADCP.load()
        ADCP['V' + beam] = xr.apply_ufunc(
            __interp1d_np,
            ADCP['Depth'] - ADCP['D' + beam],
            ADCP['VelocityBeam' + beam],
            input_core_dims=[['bin'], ['bin']],
            output_core_dims=[['gridded_bin']],
            exclude_dims={'bin'},
            vectorize=True,
            output_sizes={'gridded_bin': len(depth_offsets)},
        )
        
        ADCP['V' + beam].attrs = {"units": 'm.s-1', 'description': f'Velocity along beam {beam}, after interpolating onto isobars.'}

    ADCP = ADCP.assign_coords({'depth_offset': (['gridded_bin'], depth_offsets)})
    ADCP = ADCP.assign_coords(
        {
            'bin_depth': (
                ['time', 'gridded_bin'],
                np.tile(
                    ADCP['Depth'].values.astype('float'), (len(ADCP.gridded_bin), 1)
                ).T
                - np.tile(depth_offsets, (len(ADCP.time), 1)),
                {"units": 'm', 'description': f'Measurement depth.'}
            )
        }
    )
    return ADCP


def _rotate_BEAMS_to_XYZ(ADCP, options):
    """
    Coordinate transform that converts BEAM velocities to glider relative velocities X, Y, Z.
    Those of you with a keen eye will notice we rely on the redundancy of 4 beams to calculate 3 beam which allows you to estimate values of any missing beam from the others.
    We do this for a specific - and I promise you, rather clever - reason.
    You see, if we could calculate from 3 beams, but X would not be aligned to glider pitch.
    So either we rotate the resulting matrix, or we simply calculate from the original 4 beam matrix.
    
    TODO : Bastien Queste : double check a 3-beam algorithm gives the same result.


    Inputs
    ----------
    ADCP : xr.Dataset
        GliderAD2CP format xr.Dataset produced by the gliderad2cp.load_data() function.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function. Requires 'ADCP_mounting_direction' to be either 'top' or 'bottom'.


    Outputs
    -------
    ADCP : xr.Dataset
        Same as input but with the additional X, Y and Z velocities (X aligned with the glider, Z aligned with ADCP direction)
    
    """
    plog('Calculating X,Y,Z from isobaric 3-beam measurements.')
    a = lambda t : 1 / (2 * np.sin(t * np.pi / 180))
    b = lambda t : 1 / (4 * np.cos(t * np.pi / 180))
    c = 1  # for convex transducer head
    tf = 47.5  # theta front - Beam 1 and 3 angle from Z
    ts = 25  # theta side - Beam 2 and 4 angle from Z

    V1 = ADCP['V1'].values
    V2 = ADCP['V2'].values
    V3 = ADCP['V3'].values
    V4 = ADCP['V4'].values
    
    upcasts = (ADCP['Pitch']+options['pitch_offset']) > 0
    downcasts = ~upcasts
    
    replaced_by = lambda good_beam : (2 * b(ts) * V2 + 2 * b(ts) * V4 - 2 * b(tf) * good_beam) / (2 * b(tf))

    if options['ADCP_mounting_direction'] == 'top':
        V1[downcasts, :] = replaced_by(V3)[
            downcasts, :
        ]  # Use aft beam on downcasts when upward facing
        V3[upcasts, :] = replaced_by(V1)[
            upcasts, :
        ]  # Use fore beam on upcasts when upward facing
    else:
        V1[upcasts, :] = replaced_by(V3)[
            upcasts, :
        ]  # Use aft beam on upcasts when downward facing
        V3[downcasts, :] = replaced_by(V1)[
            downcasts, :
        ]  # Use fore beam on downcasts when downward facing

    X = c * a(tf) * V1 - c * a(tf) * V3
    Y = -c * a(ts) * V2 + c * a(ts) * V4  # TODO: sign uncertainty here
    Z = 2 * b(ts) * V2 + 2 * b(ts) * V4

    ADCP['X'] = (['time', 'gridded_bin'], X, {"units": 'm.s-1', 'description': f'Velocity along the ADCP\'s X direction.'})
    ADCP['Y'] = (['time', 'gridded_bin'], Y, {"units": 'm.s-1', 'description': f'Velocity along the ADCP\'s Y direction.'})
    ADCP['Z'] = (['time', 'gridded_bin'], Z, {"units": 'm.s-1', 'description': f'Velocity along the ADCP\'s Z direction.'})

    return ADCP


def _rotate_XYZ_to_ENU(ADCP, options):
    """
    Coordinate transform that converts velocity estimates relative to the glider (X, Y, Z) into the earth relative reference frame east, north, up (ENU)


    Inputs
    ----------
    ADCP : xr.Dataset
        GliderAD2CP format xr.Dataset produced by the gliderad2cp.load_data() function.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function. Requires 'ADCP_mounting_direction' to be either 'top' or 'bottom'.


    Outputs
    -------
    ADCP : xr.Dataset
        Same as input but with the additional X, Y and Z velocities (X aligned with the glider, Z aligned with ADCP direction)
    
    """
    def __M_xyz2enu(heading, pitch, roll):
        hh = np.pi * (heading - 90) / 180
        pp = np.pi * pitch / 180
        rr = np.pi * roll / 180

        GPT = [
            [
                np.cos(hh) * np.cos(pp),
                -np.cos(hh) * np.sin(pp) * np.sin(rr) + np.sin(hh) * np.cos(rr),
                -np.cos(hh) * np.sin(pp) * np.cos(rr) - np.sin(hh) * np.sin(rr),
            ],
            [
                -np.sin(hh) * np.cos(pp),
                np.sin(hh) * np.sin(pp) * np.sin(rr) + np.cos(hh) * np.cos(rr),
                np.sin(hh) * np.sin(pp) * np.cos(rr) - np.cos(hh) * np.sin(rr),
            ],
            [np.sin(pp), np.cos(pp) * np.sin(rr), np.cos(pp) * np.cos(rr)],
        ]
        return GPT

    if options['ADCP_mounting_direction'] == 'top':
        direction = 1
    else:
        direction = -1

    plog('Calculating E, N, U velocities from X, Y, Z velocities.')
    H = ADCP['Heading'] + options['heading_offset']
    P = ADCP['Pitch'] + options['pitch_offset']
    R = ADCP['Roll'] + options['roll_offset']
    M = __M_xyz2enu(H, P, R)

    E = (
        M[0][0] * ADCP['X']
        + M[0][1] * ADCP['Y'] * direction
        + M[0][2] * ADCP['Z'] * direction
    )
    N = (
        M[1][0] * ADCP['X']
        + M[1][1] * ADCP['Y'] * direction
        + M[1][2] * ADCP['Z'] * direction
    )
    U = (
        M[2][0] * ADCP['X']
        + M[2][1] * ADCP['Y'] * direction
        + M[2][2] * ADCP['Z'] * direction
    )

    ADCP['E'] = E
    ADCP['N'] = N
    ADCP['U'] = U
    
    plog('Calculating shear from velocities.')
    ADCP['Sh_E'] = (
        ['time', 'gridded_bin'],
        ADCP['E'].differentiate('gridded_bin').values,
    )
    ADCP['Sh_N'] = (
        ['time', 'gridded_bin'],
        ADCP['N'].differentiate('gridded_bin').values,
    )
    ADCP['Sh_U'] = (
        ['time', 'gridded_bin'],
        ADCP['U'].differentiate('gridded_bin').values,
    )
    
    ADCP['E'].attrs = {"units": 'm.s-1', 'description': f'Raw velocity along the E direction, including the glider\'s velocity through water.'}
    ADCP['N'].attrs = {"units": 'm.s-1', 'description': f'Raw velocity along the N direction, including the glider\'s velocity through water.'}
    ADCP['U'].attrs = {"units": 'm.s-1', 'description': f'Raw velocity along the U direction, including the glider\'s velocity through water.'}
    ADCP['Sh_E'].attrs = {"units": 's-1', 'description': f'Shear in the E direction.'}
    ADCP['Sh_N'].attrs = {"units": 's-1', 'description': f'Shear in the N direction.'}
    ADCP['Sh_U'].attrs = {"units": 's-1', 'description': f'Shear in the U direction.'}
    return ADCP


"""
Main
"""
def process(adcp_file_path, glider_file_path, options=None):
    """
    Calculate shear from a Nortek AD2CP mounted on a glider.
    Compatible with any glider as long as the ADCP is returning data for the 4 beams.
    

    Inputs
    ----------
    adcp_file_path : str
        Path to the AD2CP netcdf files created by the Nortek MIDAS software. Can handle wildcards through glob.
    glider_file_path : str
        Path to the glider file containing time, latitude, longitude, soundspeed, and profile_num.
        Can be csv, pandas or xarray. Can also pass pandas or xarray dataframes directly.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function.
    
        
    Outputs
    -------
    ADCP : xr.Dataset
        xr.Dataset of the AD2CP netcdf with many supplemental variables, most importantly X, Y, Z and E, N, U velocities projected onto isobars to avoid shear smearing.


    Notes
    -------
    .process() runs the following functions in this order
    1. load_data - load data from a Nortek AD2CP netCDF file and a glider data file or dataset
    2. (correct_heading - Awaiting publication, contact Bastien Queste for access)
    3. _velocity_soundspeed_correction- Corrects along beam velocity data for lack of salinity measurements in ADCP default soundspeed settings.
    4. _quality_control_velocities -  Removes bad velocity measurements based on velocity, amplitude and correlation criteria defined in options
    5. _determine_velocity_measurement_depths - Determines what depth each velocity measurement was taken at for each beam, account for glider attitude
    6. _regrid_beam_velocities_to_isobars -  Regrids beam velocities onto isobars to avoid shear smearing in the final shear data
    7. _rotate_BEAMS_to_XYZ -  Coordinate transform that converts BEAM velocities to glider relative velocities X, Y, Z
    8. _rotate_XYZ_to_ENU - Coordinate transform that converts velocity estimates relative to the glider (X, Y, Z) into the earth relative reference frame east, north, up (ENU)
    """
    # Load default options if not present.
    if not options:
        options = get_options(verbose=False)
        plog('Using default set of options. See gliderad2cp.tools.get_options() for settings.')
    
    # Load glider and ADCP data.
    # Merge necessary glider variables into the ADCP dataset.
    # Discard glider data as not necessary for other functions
    ADCP, options, _ = load_data(
                    adcp_file_path, 
                    glider_file_path, 
                    options
                    )
    
    # Correct heading based on magnetometer data.
    # Option to be implemented after publication of methods paper. Contact Bastien Queste for details.
    # if options['correct_compass_calibration']:   
    #     ADCP = correct_heading(ADCP, options)
    
    # Correct data for soundspeed and quality control based on correlation, amplitude and velocity.
    ADCP = _velocity_soundspeed_correction(ADCP)
    ADCP = _quality_control_velocities(ADCP, options)
    
    # Determine depths of velocity measurements and regrid onto isobars to prevent shear smearing.
    ADCP = _determine_velocity_measurement_depths(ADCP, options)  
    ADCP = _regrid_beam_velocities_to_isobars(ADCP, options)
    
    # Now that we have clean beam velocities, we can convert the data into XYZ and ENU.
    ADCP = _rotate_BEAMS_to_XYZ(ADCP, options)
    ADCP = _rotate_XYZ_to_ENU(ADCP, options)
    
    return ADCP


"""
gliderad2cp.tools
--------------------------
Helper funcitons for gliderad2cp
"""
import pandas as pd
import numpy as np
import logging
from datetime import datetime as dt
from scipy.interpolate import interp1d


def get_options(verbose=True, **kwargs):
    """
    Returns a dictionary containing options compatible with gliderAD2CP.
    Run with no kwargs to output a default dictionary, or with kwargs to return a full dictionary with the chose options
    """
    options = {
        'correct_compass_calibration' : [False, 'compass correction algorithm is awaiting publication and will be added upon acceptance. Contact Bastien Queste if you require.'],
        'shear_to_velocity_method' : ['integrate',],
        'ADCP_mounting_direction' : ['auto', 'top', 'bottom'],
        'QC_correlation_threshold' : [80, 'minimum acceptable along-beam correlation value.'],
        'QC_amplitude_threshold' : [80, 'maximum acceptable along-beam amplitude.'],
        'QC_velocity_threshold' : [0.8, 'maximum acceptable along-beam velocity in m.s-1.'],
        'QC_SNR_threshold' : [3, 'minimum acceptable dB above the noise floor.'],
        'velocity_regridding_distance_from_glider' : ['auto', 'array of depth-offsets from the glider, in m, at which to interpolate beam velocities onto isobars to avoid shear-smearing. Negative for bottom-mounted ADCPs.'],
        'xaxis' : [1, 'x-axis resolution in number of profiles of the final gridded products.'],
        'yaxis' : [None, 'If None: ADCP cell size. If int: y-axis resolution in metres of the final gridded products.'],
        'weight_shear_bias_regression' : [False, True, 'Give greater weight to dives with greater travel distance which can increase signal to noise.'],
        'velocity_dependent_shear_bias_correction' : [False, True, 'Determine velocity dependent shear-bias correction coefficients rather than constant coefficients.'],
        'shear_bias_regression_depth_slice' : [(0, 1000), 'A tuple containing the upper and lower depth limits over which to determine shear bias. Helpful to avoid increased noise due to surface variability. For deep diving gliders (500,1000) is good.'],
        'pitch_offset' : [0, 'value to be added to pitch to correct for transducer-compass misalignment'],
        'roll_offset' : [0, 'value to be added to roll to correct for transducer-compass misalignment'],
        'heading_offset' : [0, 'value to be added to heading to correct for transducer-compass misalignment'],
        }
        
    default = dict()
    
    if verbose:
        print('Available options are: ')
        for i, (k,v) in enumerate(options.items()):
            print('    %s : %s' % (k, v))
        print('Default setting is the first value.')
        print('Options prefaced with a minus are not yet implemented.')
        
    for i, (k,v) in enumerate(options.items()):
        default[k] = v[0]
    
    for i, (k,v) in enumerate(kwargs.items()):
        default[k] = v    
    
    return default

"""
Quality of life functions
"""
_log = logging.getLogger(__name__)
def plog(msg):
    """
    Output information to the logger with timestamp.
    """
    print(msg)
    _log.info(str(dt.now().replace(microsecond=0)) + ' : ' + str(msg))
    return None

def grid2d(x, y, v, xi=1, yi=1, fn='median'):
    """
    Quick data binning function relying on pandas.
    x,y,v are flat np.arrays of x, y coordinates and values at these points.
    xi and yi are either np.arrays of bins to be binned into, or the spacing used between min and max of x or y respectively.
    fn defines the function applied to all points from v that fall into the same bin.
    """
    if np.size(xi) == 1:
        xi = np.arange(np.nanmin(x), np.nanmax(x) + xi, xi)
    if np.size(yi) == 1:
        yi = np.arange(np.nanmin(y), np.nanmax(y) + yi, yi)

    raw = pd.DataFrame({'x': x, 'y': y, 'v': v}).dropna()

    grid = np.full([np.size(yi), np.size(xi)], np.nan)

    raw['xbins'], xbin_iter = pd.cut(raw.x, xi, retbins=True, labels=False, right=False)
    raw['ybins'], ybin_iter = pd.cut(raw.y, yi, retbins=True, labels=False, right=False)

    _tmp = raw.groupby(['xbins', 'ybins'])['v'].agg(fn)
    grid[
        _tmp.index.get_level_values(1).astype(int),
        _tmp.index.get_level_values(0).astype(int),
    ] = _tmp.values

    XI, YI = np.meshgrid(xi, yi, indexing='ij')
    return grid, XI.T, YI.T

def interp(x, y, xi):
    """
    Interpolation shortcut that removes NaN data and ignores boundary errors.
    """
    _gg = np.isfinite(x + y)
    return interp1d(x[_gg], y[_gg], bounds_error=False, fill_value=np.nan)(xi)


"""
Calculate depth-resolved currents from shear data calculated by the gliderAD2CP.process_shear.process() function.
This functionality can handle loiter dives and irregular profiles but it requires that your ADCP is 
always on so that DAC is representative of the baroclinic profile collected by the ADCP.


gliderad2cp.process_currents
------------------------------
process
    The main function which converts shear data from gliderAD2CP.process_shear.process() to referenced velocities.
get_DAC
    Calculate dive-averaged currents using the ADCP as a DVL.
_grid_shear
    Grid shear according to specifications.
_grid_velocity
    Assemble shear measurements to reconstruct velocity profiles.
_reference_velocity
    Reference vertical velocity profiles to dive-averaged currents, paying attenting to the time spent at each depth.
    
Notes
-------
.process() runs the following functions in this order

1. get_DAC - Calculate dive-averaged currents using the ADCP as a DVL.
2. _grid_shear - Grid shear according to specifications.
3. _grid_velocity - Assemble shear measurements to reconstruct velocity profiles.
4. _reference_velocity - Reference vertical velocity profiles to dive-averaged currents, paying attenting to the time spent at each depth.

(Optional steps:) - Not coded up yet

5. assess_shear_bias - Not coded up yet
6. correct_shear_bias - Not coded up yet.

"""

import numpy as np
import pandas as pd
import warnings
import xarray as xr
import gsw
from scipy.interpolate import interp1d
from .tools import plog, grid2d, interp, get_options

warnings.filterwarnings(action="ignore", message="Mean of empty slice")
warnings.filterwarnings(action="ignore", message="invalid value encountered in divide")
warnings.filterwarnings(action="ignore", message="invalid value encountered in true_divide")
warnings.filterwarnings(action="ignore", message="Degrees of freedom <= 0 for slice.")
warnings.filterwarnings(action="ignore", message="All-NaN slice encountered")
warnings.simplefilter(action='ignore', category=FutureWarning)


"""
Processing functions
"""
def get_DAC(ADCP, gps_predive, gps_postdive):
    """
    Calculate dive-averaged currents using the ADCP as a DVL.


    Inputs
    ----------
    ADCP : xarray.Dataframe
        Output from the gliderAD2CP.process_shear() function.
    gps_predive : numpy.array
        Mx3 numpy array where M in the number of dives performed by the glider.
        Columns are np.datetime64[ns], longitude and latitude.
    gps_postdive : numpy.array
        Mx3 numpy array where M in the number of dives performed by the glider. 
        Columns are np.datetime64[ns], longitude and latitude.
         
    Outputs
    -------
    DAC : dictionary containing : 
        dac
            Mx2 np.array containing eastward DAC (column 0) and northward DAC (column 1), for M dives identified.
        dive_time
            Mx1 np.array containing mean time for the dive in np.datetime64[ns] format.
        dive_duration
            Mx1 np.array containing mean duration for the dive in seconds.
        dxy_gps, dxy_dvl
            Mx2 np.arrays containing eastward (col 0) and northward (col 1) displacement in m,
            calculated from GPS and from ADCP as a DVL respectively.
    
    """
    plog('Calculating dive-averaged currents from dive and surfacing GPS coordinates.')
    ## Process time data
    time = ADCP.time.values.astype("float") / 1e9
    deltat = np.append(np.diff(time), np.nan)
    f = np.nanmedian(deltat)
    deltat[deltat > f*10] = np.nan
    deltat = np.nan_to_num(deltat)
    
    ## Create interpolannt to query DVL travel at time t.
    travel_e = interp1d(
            time,
            np.cumsum(np.nan_to_num(-(ADCP["E"].mean(dim="gridded_bin") * deltat).values)),
            bounds_error=False, fill_value=np.nan
            )
    travel_n = interp1d(
            time,
            np.cumsum(np.nan_to_num(-(ADCP["N"].mean(dim="gridded_bin") * deltat).values)),
            bounds_error=False, fill_value=np.nan
            )
    
    ## Preallocate arrays
    r,c = np.shape(gps_predive)
    
    dive_time = np.full(r, np.nan) # Duration in seconds
    dive_duration = np.full(r, np.nan) # Duration in seconds
    
    dac = np.full([r,2], np.nan) # dive-averaged currents in m.s-1
    dxy_gps =  np.full([r,2], np.nan) # x/y displacement in m, over earth, for the dive calculated from GPS coordinates
    dxy_dvl =  np.full([r,2], np.nan) # x/y displacement in m, through water, for the dive calculated from ADCP as DVL
    
    ## Useful tidbits
    pre_t = gps_predive[:, 0].astype('datetime64[s]').astype(np.float64) # s
    post_t = gps_postdive[:, 0].astype('datetime64[s]').astype(np.float64) # s

    lon2m = lambda x, y : gsw.distance([x-0.0005, x + 0.0005], [y, y])*1000
    lat2m = lambda x, y : gsw.distance([x, x], [y-0.0005, y+0.0005])*1000
    
    ## Calculate displacements
    for idx in range(len(gps_predive)):
        dt = post_t - pre_t[idx] # s
        dt[dt < 0] = np.nan
        
        if all(np.isnan(dt)):
            plog(f'    Could not find a surfacing after dive starting: {gps_predive[idx,0]}.')
            continue
        
        idy = np.nanargmin(dt)
        dive_duration[idx] = dt[idy] # s
        dive_time[idx] = (post_t[idy]+pre_t[idx])/2
        
        dxy_dvl[idx,0] = travel_e(post_t[idy]) - travel_e(pre_t[idx])
        dxy_dvl[idx,1] = travel_n(post_t[idy]) - travel_n(pre_t[idx])
        
        dxy_gps[idx,0] = (gps_postdive[idy,1] - gps_predive[idx,1]) * lon2m(gps_predive[idx,1],gps_predive[idx,2])
        dxy_gps[idx,1] = (gps_postdive[idy,2] - gps_predive[idx,2]) * lat2m(gps_predive[idx,1],gps_predive[idx,2])
    
    ## Calculate dive-averaged currents
    dac = (dxy_gps - dxy_dvl) / np.vstack([dive_duration,dive_duration]).T
    dive_time = pd.to_datetime(dive_time*1e9).values
    
    ## Create output dictionary
    DAC = {'dac': dac, 'dive_time': dive_time, 'dive_duration': dive_duration, 'xy_displacement_over_land': dxy_gps ,'xy_displacement_through_water': dxy_dvl}
    
    return DAC

def _grid_shear(ADCP, options, xi, yi):
    """
    Grid shear according to specifications.


    Inputs
    ----------
    ADCP : xarray.Dataframe
        Output from the gliderAD2CP.process_shear() function.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function.
    xi : numpy.array
        Array of bins in x-axis.
    yi : numpy.array
        Array of bins in y-axis.
         
    Outputs
    -------
    currents : xr.Dataset
        Dataset containing gridded shear and various statistical metrics, as well as time spent per bin by the glider.    
        
    """
    
    plog(f'Gridding shear data.')
    plog(f'    Creating output structure and time-based metrics.')
    
    # Create structure
    currents = xr.Dataset(coords={"depth": yi, "profile_index": xi})
    
    currents["depth"].attrs = {"units": 'm', 'description': 'Central measurement depth in meters.'}
    currents["profile_index"].attrs = {"units": '', 'description': 'Central profile number of measurement.'}
    
    
    # Get coords for gridding function below in the right dimensions
    x, y = xr.broadcast(ADCP.profile_number, ADCP.bin_depth)
    t, _ = xr.broadcast(ADCP.time, ADCP.bin_depth)
    x = x.values.flatten()
    y = y.values.flatten()
    t = t.values.astype('float').flatten()

    # Create time coordinates
    time = np.nanmedian(grid2d(x, y, t, xi=xi, yi=yi, fn=np.nanmedian)[0], axis=0)
    currents = currents.assign_coords(time=("profile_index", pd.to_datetime(time)))
    
    # Calculate time in bin
    # This metric is important when referencing to DAC later.
    # We weight the baroclinic velocity by time spent in each bin when referencing to DAC.
    # This is relevant on loiter dives for example, where we may spend 24 hours loitering 
    # and as a result the DAC is heavily biased to the velocity at the loiter depth.
    # Calculating time in each bin is very easy (differentiate sampling frequency and sum in bins)
    # ... this works great to calculate loiters etc.... 
    # ... BUT we have very long periods at the surface which are NOT included in the DAC.
    # Here we set durations to 0 when dP/dt is under a threshold near the surface.
    # It's not perfect, but it's fairly cross-compatible assumption.    
    duration = np.gradient(ADCP.time.values.astype('float'))/1e9 # Time spent at that depth
    # Set surface values to 0. Imperfect because pressure sensor drift but captures a good amout. We lose some good data but very minimal.
    duration[ADCP.Depth.values < 1] = 0 
    # Same with a looser depth threshold but only if speeds are less than 4cm/s.
    duration[
        (ADCP.Depth.values < 3) & (np.abs(np.gradient(ADCP.Depth.values, ADCP.time.values.astype('float')/1e9)) < 0.04)
         ] = 0
    currents['time_in_bin'] = (('depth', 'profile_index'),
                               grid2d(ADCP.profile_number.values, ADCP.Depth.values, 
                                 duration,
                                 xi=xi, yi=yi, fn=np.nansum)[0]
                              )
    currents['time_in_bin'].attrs = {"units": 's', 'description': 'Cumulative time spent by the glider in this depth x profile bin.'}

    # Gridding function for shear
    plog(f'    Gridding flight metrics.')
    grid = lambda var_name, fn_name : grid2d(x, y, ADCP[var_name].values.flatten(), xi=xi, yi=yi, fn=fn_name)[0]
    count_valid = lambda x : np.count_nonzero(np.isfinite(x))
    cos_mean = lambda head : np.nanmean(np.cos(np.deg2rad(head))) # N
    sin_mean = lambda head : np.nanmean(np.sin(np.deg2rad(head))) # E
    
    currents['heading_N'] = (('depth', 'profile_index'),
                       grid2d(ADCP.profile_number.values, ADCP.Depth.values, 
                         ADCP.Heading.values + options['heading_offset'],
                         xi=xi, yi=yi, fn=cos_mean)[0],
                       {"units": '', 'description': 'N-ward cartesian component of heading.'}
                      )
    currents['heading_E'] = (('depth', 'profile_index'),
                           grid2d(ADCP.profile_number.values, ADCP.Depth.values, 
                             ADCP.Heading.values + options['heading_offset'],
                             xi=xi, yi=yi, fn=sin_mean)[0],
                           {"units": '', 'description': 'E-ward cartesian component of heading.'}
                          )
    currents['speed_through_water'] = (('depth', 'profile_index'),
                           grid2d(ADCP.profile_number.values, ADCP.Depth.values, 
                             np.sqrt(ADCP.X.mean(dim='gridded_bin')**2 +ADCP.Y.mean(dim='gridded_bin')**2 +ADCP.Z.mean(dim='gridded_bin')**2),
                             xi=xi, yi=yi, fn=np.nanmean)[0],
                           {"units": 'm.s-1', 'description': 'Glider speed through water, estimated using the ADCP as a doppler velocity logger.'}
                          )
    
    # Grid shear data and metrics
    plog(f'    Gridding eastward shear metrics.')
    currents['shear_E_mean'] = (('depth', 'profile_index'),   grid('Sh_E', np.nanmean),     {"units": 's-1', 'description': 'Mean shear in the E-direction.'})
    currents['shear_E_median'] = (('depth', 'profile_index'), grid('Sh_E', np.nanmedian),   {"units": 's-1', 'description': 'Median shear in the E-direction.'})
    currents['shear_E_stddev'] = (('depth', 'profile_index'), grid('Sh_E', np.nanstd),      {"units": 's-1', 'description': 'Std. dev. of shear in the E-direction.'})
    currents['shear_E_count'] = (('depth', 'profile_index'),  grid('Sh_E', count_valid),    {"units": '', 'description': 'Number of valid shear measurements in the E-direction.'})

    plog(f'    Gridding northward shear metrics.')
    currents['shear_N_mean'] = (('depth', 'profile_index'),   grid('Sh_N', np.nanmean),     {"units": 's-1', 'description': 'Mean shear in the N-direction.'})
    currents['shear_N_median'] = (('depth', 'profile_index'), grid('Sh_N', np.nanmedian),   {"units": 's-1', 'description': 'Median shear in the N-direction.'})
    currents['shear_N_stddev'] = (('depth', 'profile_index'), grid('Sh_N', np.nanstd),      {"units": 's-1', 'description': 'Std. dev. of shear in the N-direction.'})
    currents['shear_N_count'] = (('depth', 'profile_index'),  grid('Sh_N', count_valid),    {"units": '', 'description': 'Number of valid shear measurements in the N-direction.'})
    
    return currents

def _grid_velocity(currents, method='integrate'):
    """
    Assemble shear measurements to reconstruct velocity profiles.
    Currently only able to integrate in the vertical, but I have aspirations to add the least-squared method too.
    Untested and likely not compatible with irregular grids.


    Inputs
    ----------
    currents : xr.Dataset
        Dataset containing gridded shear produced by the process_currents.grid_shear() function.
    method : str, default = 'integrate'
        Which method to use to reconstruct velocity profiles. Currently only "integrate" is available.
        "lsq" coming one day. maybe.
        
         
    Outputs
    -------
    currents : xr.Dataset
        Dataset containing gridded shear, statistical metrics, time spent per bin by the glider, and unreferenced velocity profiles.
        
    """
    def __integrate():
        def __integrate_calc(Sh):
            # Integrate shear vertically
            _bd = ~np.isfinite(Sh)  # Preserve what are originally NaN values to recover later as need conversion to 0 for cumsum-
            Sh = np.nan_to_num(Sh)  # Replace NaNs with 0 for cumsum
            y_res = np.gradient(currents.depth.values)
            y_res = np.broadcast_to(y_res[:,np.newaxis],Sh.shape)
            V = np.cumsum(Sh * y_res, axis=0)  # Cumulative sum of shear to recover velocity profile
            V[_bd] = np.nan  # Return NaNs to their rightful place.
            V = V - np.tile(
                np.nanmean(V, axis=0), (np.shape(V)[0], 1)
            )  # Make mean of baroclinic profiles equal to 0
            return V        
        currents['velocity_E_no_reference'] = (('depth', 'profile_index'), __integrate_calc(currents.shear_E_median.values), {"units": 'm.s-1', 'description': f'Unreferenced velocity profile in the E direction.'})
        currents['velocity_N_no_reference'] = (('depth', 'profile_index'), __integrate_calc(currents.shear_N_median.values), {"units": 'm.s-1', 'description': f'Unreferenced velocity profile in the N direction.'})
    def __lsq():
        # Space for Martin Visbeck's least squared method
        return False
    
    plog(f'Calculating unreferenced velocity profiles from shear.')
    
    if method == 'integrate':
        fn = __integrate
    elif method == 'lsq':
        plog('WARNING: least squared method not yet coded up. Oops. Any volunteers?')
        fn = __lsq
        
    fn()
    
    return currents

def _reference_velocity(currents, DAC):
    """
    Reference vertical velocity profiles to dive-averaged currents, paying attention to the time spent at each depth.
    This ensures that dives with irregular vertical speeds or periods of loitering at depth are still reproduced correctly.
    As a result, one cannot expect the *depth*-averaged velocity of a dive to match the *dive*-averaged currents.


    Inputs
    ----------
    currents : xr.Dataset
        Dataset containing gridded shear produced by the process_currents.grid_shear() function.
    DAC : dictionary containing fields :
        dac : 
            Mx2 np.array containing eastward DAC (column 0) and northward DAC (column 1), for M dives identified.
        dive_time : 
            Mx1 np.array containing mean time for the dive in np.datetime64[ns] format.
        
         
    Outputs
    -------
    currents : xr.Dataset
        Dataset containing gridded shear, statistical metrics, time spent per bin by the glider, and unreferenced velocity profiles.
        
    """
    
    plog('Referencing velocity profiles to dive-averaged currents using baroclinic velocity weighting.')
    # DAC = \int_{0}^{z} \frac{velocity * time spent in bin}{time spent in profile} + referencing
    # DAC - \int_{0}^{z} \frac{velocity * time spent in bin}{time spent in profile} = referencing
    
    time_in_bin, depth = xr.broadcast(currents.time_in_bin, currents.depth)
    time_in_bin = time_in_bin.values
    depth = depth.values
    time_in_bin_mean = np.nanmean(time_in_bin, axis=0)
    
    
    for DAC_idx, direction in enumerate(['E','N']):
        currents[f'DAC_{direction}'] = ('profile_index', interp(DAC['dive_time'].astype('float'), DAC['dac'][:,DAC_idx], currents.time.values.astype('float')))
        currents[f'displacement_though_water_{direction}'] = ('profile_index', interp(DAC['dive_time'].astype('float'), DAC['xy_displacement_through_water'][:,DAC_idx], currents.time.values.astype('float')))

        
        velocity = currents[f'velocity_{direction}_no_reference'].values * time_in_bin / np.broadcast_to(time_in_bin_mean[np.newaxis,:],time_in_bin.shape)
        reference = currents[f'DAC_{direction}'].values - np.nanmean(velocity, axis=0)

        currents[f'reference_{direction}_from_DAC'] = ('profile_index', reference)

        currents[f'velocity_{direction}_DAC_reference'] = currents[f'velocity_{direction}_no_reference'] + currents[f'reference_{direction}_from_DAC']
        
        
        currents[f'DAC_{direction}'].attrs = {"units": 'm.s-1', 'description': f'Dive-averaged currents in the {direction} direction.'}
        currents[f'displacement_though_water_{direction}'].attrs = {"units": 'm', 'description': f'Glider trajectory through water in the {direction} direction since mission start.'}
        currents[f'reference_{direction}_from_DAC'].attrs = {"units": 'm.s-1', 'description': f'Reference velocity to correct unreferenced velocity profile; weighted by time-in-bin to account for loiters and uneven dives. In the {direction} direction.'}
        currents[f'velocity_{direction}_DAC_reference'].attrs = {"units": 'm.s-1', 'description': f'DAC-referenced velocity profile in the {direction} direction.'}

    return currents


"""
Main
"""
def process(ADCP, gps_predive, gps_postdive, options=None):
    """
    Calculate depth-resolved currents from shear data calculated by the gliderAD2CP.process_shear.process() function.
    This functionality can handle loiter dives and irregular profiles but it requires that your ADCP is 
    always on so that DAC is representative of the baroclinic profile collected by the ADCP.
    

    Inputs
    ----------
    ADCP : xarray.Dataframe
        Output from the gliderAD2CP.process_shear() function.
    gps_predive : numpy.array
        Mx3 numpy array where M in the number of dives performed by the glider.
        Columns are np.datetime64[ns], longitude and latitude.
    gps_postdive : numpy.array
        Mx3 numpy array where M in the number of dives performed by the glider. 
        Columns are np.datetime64[ns], longitude and latitude.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function.
    
        
    Outputs
    -------
    currents : xr.Dataset
        Dataset containing gridded shear, statistical metrics, time spent per bin by the glider, unreferenced velocity profiles and DAC-referenced velocities.

    Notes
    ------
    .process() runs the following functions in this order

    1. get_DAC - Calculate dive-averaged currents using the ADCP as a DVL.
    2. _grid_shear - Grid shear according to specifications.
    3. _grid_velocity - Assemble shear measurements to reconstruct velocity profiles.
    4. _reference_velocity - Reference vertical velocity profiles to dive-averaged currents, paying attention to the time spent at each depth.
    
    """
    # Load default options if not present.
    if not options:
        options = get_options(verbose=False)
        plog('Using default set of options. See gliderad2cp.tools.get_options() for settings.')
    
    xi = options['xaxis']
    yi = options['yaxis']
    
    if yi is None:
        yi = ADCP.attrs['avg_cellSize'].astype('float')
        plog(f'Setting y-axis resolution to {yi}, based on ADCP sampling cell size.')
        
    if np.size(xi) == 1:
        xi = np.arange(np.nanmin(ADCP.profile_number.values), np.nanmax(ADCP.profile_number.values) + xi, xi)
    if np.size(yi) == 1:
        yi = np.arange(0, np.nanmax(np.nanmax(ADCP.bin_depth)) + yi, yi)
        
    DAC = get_DAC(ADCP, gps_predive, gps_postdive)
    
    currents = _grid_shear(ADCP, options, xi, yi)
    currents = _grid_velocity(currents, method=options['shear_to_velocity_method'])
    currents = _reference_velocity(currents,DAC)
    
    return currents, DAC


"""
gliderad2cp.download_example_data
-----------------------------------
Functions for downloading sample datasets
"""

import pooch
server = "https://erddap.observations.voiceoftheocean.org/examples/gliderad2cp/"
data_source = pooch.create(
    path=pooch.os_cache("gliderad2cp"),
    base_url=server,
    registry={
        "adcp_profiles_160_to_210.nc": "sha256:ee74a61af86547be0ea40d5bc6e09b88fdf89f87c4e48940a8f9fc694c2717b5",
        "glider_profiles_160_to_210.pqt": "sha256:86b63e82a2cbc9128fc015299de09d8090344ff01614d9d9f96a8712a30e63a7",
        "sea055_M82.nc": "sha256:55f033a6c14a7e1882e8ed981e3dc39ec4efade4a269eb16675fbbeeced826fc",
        "sea055_M82.ad2cp.00000.nc": "sha256:8affe4520133aba6336c4db90cbbe5264fa10db49c21887f921f680d0d6cabe4",
    },
)

def load_sample_dataset(dataset_name="SEA055_M82.nc"):
    """Download sample datasets for use with gliderad2cp

    Parameters
    ----------
    dataset_name: str, optional
        Default is "sea045_20230530T0832_delayed.nc".

    Raises
    ------
    ValueError: 
        If the requests dataset is not known, raises a value error

    Returns
    -------
    xarray.Dataset: 
        Requested sample dataset
    """
    if dataset_name in data_source.registry.keys():
        file_path = data_source.fetch(dataset_name)
        return file_path
    else:
        msg = f"Requested sample dataset {dataset_name} not known. Specify one of the following available datasets: {list(data_source.registry.keys())}"
        raise KeyError(msg)



"""
Correct velocity profiles obtained by ADCP gliders for shear-bias (as described in Todd et al., 2017; Sec.3.b.2 https://doi.org/10.1175/JTECH-D-16-0156.1)
Applies a different approach to Todd et al. in that we only utilise variance at depth as validation.
Instead, we aim to reduce the slope of the linear regression :

profile_mean(shear - shear_bias_correction) = profile_mean_naturally_occuring_shear + slope * displacement_through_water.

We determine two separate shear_bias_correction coefficients, one in the along-glider direction and one in the across glider direction.
This is akin to subtracting a portion of the glider's travel through water.

A second (potentially better) correction is also coded up which subtracts a shear_bias_correction which scales with glider speed. 
Preliminary work indicates that shear bias is instrument-velocity dependent, but until this is proven/published, this correction remains the
non-default correction. NB: the author of this code uses this one all the time.


gliderad2cp.process_bias
--------------------------
process
    The main function which will run through the whole shear bias correction automatically for you.
visualise
    Function outputting two plots which can be used to assess shear bias and its correction.
regress_bias
    Uses a minimisation function to identify optimal correction parameters
correct_bias
    Outputs DAC-referenced velocities that have been corrected for shear bias.
__linear_regression
    A weighted, linear least squares regression function with optional plotting and indexing capability.

Notes
-------
.process() runs the following functions in this order

1. visualise
2. regress_bias
3. correct_bias
4. visualise

"""

import warnings
import numpy as np
from scipy.stats import t
from scipy.optimize import fmin 
import matplotlib.pyplot as plt
from .tools import plog

warnings.filterwarnings(action="ignore", message="Mean of empty slice")
warnings.filterwarnings(action="ignore", message="invalid value encountered in divide")
warnings.filterwarnings(action="ignore", message="invalid value encountered in true_divide")
warnings.filterwarnings(action="ignore", message="Degrees of freedom <= 0 for slice.")

def _linear_regression(x,y,options,c=None,mask=True,plot=False):
    """
    Perform an optionally-weighted linear least squares regression on x and y, providing 95% confidence intervals and optionally plotting the data.

    Inputs
    ----------
    x,y,c : np.array
        Arrays containing x and y values for the linear regression, and an optional color argument for the scatter plot.
    mask : bool array
        Boolean array of same size as x, indicating which values to keep in the regression.
    plot : bool
        If true, produce a plot of the linear regression.
        
    Outputs
    -------
    slope : 
        Returns the slope of the linear regression
    
    """
    _gd = np.isfinite(x+y) & mask & (np.abs(y) < 0.01) & (np.abs(x) > 100)
    x = np.reshape(x[_gd],[-1,1])
    y = np.reshape(y[_gd],[-1,1])
    
    if options['weight_shear_bias_regression']:
        weights = x.flatten() # Weight by distance travelled, helps with weird shear spikes present on short/shallow dives
    else:
        weights = np.ones(len(x)) # Define equal weights
    
    # Add an intercept term (column of 1s) to X
    X = np.hstack( [np.ones((x.shape[0], 1)), x] )

    # Define weights (example: higher weight for larger x values)
    # weights = np.ones(len(x)) # Define weights for each observation
    W = np.diag(weights) # Create a diagonal matrix from weights

    # Calculate coefficients using the weighted normal equation
    # β = (X^T W X)^(-1) X^T W y
    X_T = X.T # Transpose of X
    beta = np.linalg.inv(X_T @ W @ X) @ (X_T @ W @ y) # Solve for β
    
    # Calculate residuals and residual variance
    y_pred = X @ beta  # Predicted values
    residuals = y - y_pred # Residuals
    n, p = X.shape # Number of observations, parameters
    sigma_squared = (residuals.T @ residuals) / (n - p) # Residual variance

    # Variance-Covariance Matrix
    var_cov_matrix = sigma_squared * np.linalg.inv(X_T @ X)

    # Extract standard error for the slope (2nd coefficient)
    slope_se = np.sqrt(var_cov_matrix[1, 1]) # Standard error of the slope
    intercept_se = np.sqrt(var_cov_matrix[0, 0]) # Standard error of the slope

    # Calculate 95% confidence interval
    t_critical = t.ppf(0.975, df=n-p) # Two-tailed t critical value
    
    intercept = beta[0][0].astype('float')
    slope = beta[1][0].astype('float')
    
    slope_ci = (slope - t_critical * slope_se, slope + t_critical * slope_se)
    intercept_ci = (intercept - t_critical * intercept_se, intercept + t_critical * intercept_se)
    
    
    if plot:
        if c is None:
            plt.plot(x, y, '.', color='k', markersize=1, label='Regressed data')
        else:
            plt.scatter(x, y, 3, c, label='Regressed data')
            plt.colorbar()
        plt.plot(x, intercept_ci[0] + slope_ci[0]*x, 'r', alpha=0.5, linestyle=':')
        plt.plot(x, intercept_ci[0] + slope_ci[1]*x, 'r', alpha=0.5, linestyle=':')
        plt.plot(x, intercept_ci[1] + slope_ci[0]*x, 'r', alpha=0.5, linestyle=':')
        plt.plot(x, intercept_ci[1] + slope_ci[1]*x, 'r', alpha=0.5, linestyle=':')
        
        plt.plot(x, intercept + slope*x, 'r', label='Regression')
        plt.legend()
        
        color = 'lightgrey'
        if np.abs(3*t_critical*slope_se) < np.abs(slope):
            color = 'orange'
        if np.abs(5*t_critical*slope_se) < np.abs(slope):
            color = 'red'
        plt.title(f"slope (95%): {slope:.2E} +/- {t_critical*slope_se:.2E}", color=color)

    return slope


def visualise(currents, options, plot_all=True):
    """
    Produce collection of plots used to estimate shear bias.


    Inputs
    ----------
    currents : xr.Dataset
        Dataset containing gridded shear and velocity data produced by the gliderad2cp.process_currents functions.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function.
        
        
    Outputs
    -------
    None
    
    """
    corr_present = False
    if 'velocity_E_DAC_reference_sb_corrected' in currents: # Assume N also present
        corr_present = True
    
    upper,lower = options['shear_bias_regression_depth_slice']
    # Plot linear regression of shear versus displacement
    plt.figure(figsize=(10,20))
    plt.tight_layout()
    directions = ['N','E']
    idx = 0
    for l1 in directions:
        for l2 in directions:
            idx += 1
            plt.subplot(4,2,idx)
            if corr_present:
                _linear_regression(
                    currents[f'displacement_though_water_{l1}'].values, 
                    (currents[f'velocity_{l2}_DAC_reference_sb_corrected']).sel(depth=slice(upper,lower)).differentiate('depth').mean(dim='depth', skipna=True).values,
                    options,
                    plot=True
                )
            else:
                _linear_regression(
                    currents[f'displacement_though_water_{l1}'].values, 
                    (currents[f'velocity_{l2}_no_reference']).sel(depth=slice(upper,lower)).differentiate('depth').mean(dim='depth', skipna=True).values,
                    options,
                    plot=True
                )
            plt.xlabel(f'Displacement {l1} per profile')
            plt.ylabel(f'Mean shear {l2} per profile')
    
    if plot_all:
        # Plots of referenced velocity variance with depth
        plt.subplot(2,2,4)
        plt.plot(currents.velocity_E_DAC_reference.var(dim='profile_index'), currents.depth, ':r', label='E$_{orig}$')
        plt.plot(currents.velocity_N_DAC_reference.var(dim='profile_index'), currents.depth, ':b', label='N$_{orig}$')
        if corr_present:
            plt.plot(currents.velocity_E_DAC_reference_sb_corrected.var(dim='profile_index'), currents.depth, '-r', label='E$_{corr}$')
            plt.plot(currents.velocity_N_DAC_reference_sb_corrected.var(dim='profile_index'), currents.depth, '-b', label='N$_{corr}$')
        plt.legend()
        plt.xlabel('Velocity Variance')
        plt.ylabel('Depth')
        plt.gca().invert_yaxis()
        YL = plt.ylim()
        
        cl = np.max(np.abs([currents.velocity_E_DAC_reference.max(),currents.velocity_E_DAC_reference.min(),currents.velocity_N_DAC_reference.max(),currents.velocity_N_DAC_reference.min()]))
        cl = np.array([-1,1])*cl
        
        # Velocity sections
        plt.subplot(8,2,9)
        plt.set_cmap('bwr')
        plt.pcolormesh(currents.profile_index, currents.depth, currents.velocity_E_DAC_reference)
        plt.colorbar(label='E$_{orig}$')
        plt.clim(cl)
        XL = plt.xlim()
        plt.ylim(YL)

        if corr_present:
            plt.subplot(8,2,11)
            plt.set_cmap('bwr')
            plt.pcolormesh(currents.profile_index, currents.depth, currents.velocity_E_DAC_reference_sb_corrected)
            plt.colorbar(label='E$_{corr}$')
            plt.clim(cl)
            plt.xlim(XL)
            plt.ylim(YL)

        plt.subplot(8,2,13)
        plt.set_cmap('bwr')
        plt.pcolormesh(currents.profile_index, currents.depth, currents.velocity_N_DAC_reference)
        plt.colorbar(label='N$_{orig}$')
        plt.clim(cl)
        plt.xlim(XL)
        plt.ylim(YL)

        if corr_present:
            plt.subplot(8,2,15)
            plt.set_cmap('bwr')
            plt.pcolormesh(currents.profile_index, currents.depth, currents.velocity_N_DAC_reference_sb_corrected)
            plt.colorbar(label='N$_{corr}$')
            plt.xlim(XL)
            plt.ylim(YL)
            plt.clim(cl)
    
    plt.tight_layout()
    
    return None

def regress_bias(currents,options):
    """
    Determines shear bias correction coefficients by empirically minimimising the slope of various linear regressions.
    Linear regressions are performed on displacement through water and mean shear of the water column in combination of directions.
    

    Inputs
    ----------
    currents : xr.Dataset
        Dataset containing gridded shear and velocity data produced by the gliderad2cp.process_currents functions.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function.
        
        
    Outputs
    -------
    bias_along_glider, bias_across_glider : float
        Floats which define the scaling factor applied to the shear bias correction in the along and across glider directions.
        
    """

    def __score(coeffs, *args):
        currs = correct_bias(args[0], args[1], coeffs[0], coeffs[1])
        directions = ['N','E']
        score = 0.
        idx = 0
        
        upper,lower = options['shear_bias_regression_depth_slice']
        for l1 in directions:
            for l2 in directions:
                idx += 1
                score = score + np.abs(_linear_regression(
                        currs[f'displacement_though_water_{l1}'].values, 
                        currs[f'velocity_{l2}_DAC_reference_sb_corrected'].sel(depth=slice(upper,lower)).differentiate('depth').mean(dim='depth', skipna=True).values,
                        options,
                    ))*1e9
        return score
    
    results = fmin(__score, [0,0], args=(currents,options), maxiter=300, disp=True, xtol=1e-9, ftol=1e-9)
    
    plog(f'Final results of shear bias regression: ')
    plog(f'    Along-glider bias :  {results[0]:.2E}')
    plog(f'    Across-glider bias : {results[1]:.2E}')
    
    return results[0], results[1]

    
def correct_bias(currents, options, bias_along_glider, bias_across_glider):
    """
    Calculates the artificial velocity profile created by the shear bias and outputs two new variables:
        "velocity_{direction}_DAC_reference_sb_corrected" - DAC referenced, shear-bias corrected velocity profiles.
        "shear_bias_velocity_{direction}" - the velocity profile error induced by the shear bias.
    

    Inputs
    ----------
    currents : xr.Dataset
        Dataset containing gridded shear and velocity data produced by the gliderad2cp.process_currents functions.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function.
    bias_along_glider, bias_across_glider : float
        Floats which define the scaling factor applied to the shear bias correction in the along and across glider directions.
        
        
    Outputs
    -------
    currents : xr.Dataset
        Dataset containing gridded shear, statistical metrics, time spent per bin by the glider, and unreferenced velocity profiles.
        
    """
    mask = np.isfinite(currents[f'velocity_E_DAC_reference'].values)
    mask[mask == False] = np.nan
    
    if options['velocity_dependent_shear_bias_correction']:
        currents['shear_bias_velocity_E'] =  (currents.speed_through_water * (currents.heading_E*bias_along_glider - currents.heading_N*bias_across_glider) * currents.depth.differentiate('depth')).cumsum('depth') * mask
        currents['shear_bias_velocity_N'] =  (currents.speed_through_water * (currents.heading_N*bias_along_glider - currents.heading_E*bias_across_glider) * currents.depth.differentiate('depth')).cumsum('depth') * mask
    else:
        currents['shear_bias_velocity_E'] =  ((currents.heading_E*bias_along_glider - currents.heading_N*bias_across_glider) * currents.depth.differentiate('depth')).cumsum('depth') * mask
        currents['shear_bias_velocity_N'] =  ((currents.heading_N*bias_along_glider - currents.heading_E*bias_across_glider) * currents.depth.differentiate('depth')).cumsum('depth') * mask
    
    directions = ['N','E']
    
    for l in directions:
        currents[f'shear_bias_velocity_{l}'] = currents[f'shear_bias_velocity_{l}'] - currents[f'shear_bias_velocity_{l}'].mean(dim='depth')        
        currents[f'velocity_{l}_DAC_reference_sb_corrected'] = currents[f'velocity_{l}_DAC_reference'] - currents[f'shear_bias_velocity_{l}']
    
        currents[f'shear_bias_velocity_{l}'].attrs = {"units": 'm.s-1', 'description': f'Erroneous velocity induced by shear-bias in the {l} direction.'}
        currents[f'velocity_{l}_DAC_reference_sb_corrected'].attrs = {"units": 'm.s-1', 'description': f'DAC-referenced and shear-bias corrected, velocity profile in the {l} direction.'}
    
    return currents

def process(currents, options):
    """
    Correct velocity profiles obtained by ADCP gliders for shear-bias (as described in Todd et al., 2017; Sec.3.b.2 https://doi.org/10.1175/JTECH-D-16-0156.1)
    Applies a different approach to Todd et al. in that we only utilise variance at depth as validation.
    Instead, we aim to reduce the slope of the linear regression :

    profile_mean(shear - shear_bias_correction) = profile_mean_naturally_occuring_shear + slope * displacement_through_water.

    We determine two separate shear_bias_correction coefficients, one in the along-glider direction and one in the across glider direction.
    This is akin to subtracting a portion of the glider's travel through water.

    A second (potentially better) correction is also coded up which subtracts a shear_bias_correction which scales with glider speed. 
    Preliminary work indicates that shear bias is instrument-velocity dependent, but until this is proven/published, this correction remains the
    non-default correction. NB: the author of this code uses this one all the time.
    

    Inputs
    ----------
    currents : xr.Dataset
        Dataset containing gridded shear and velocity data produced by the gliderad2cp.process_currents functions.
    options : dict
        Set of options for gliderAD2CP, created by the gliderad2cp.tools.get_options() function.
        
        
    Outputs
    -------
    currents : xr.Dataset
        Dataset containing gridded shear, statistical metrics, time spent per bin by the glider, and unreferenced velocity profiles.
        

    Notes
    -------
    .process() runs the following functions in this order
    
    1. visualise -  Produce collection of plots used to visalise estimated shear bias
    2. regress_bias - Determine shear bias correction coefficients by empirically minimimising the slope of various linear regressions
    3. correct_bias -  Calculate the artificial velocity profile created by the shear bias and correct it in a new variable
    4. visualise - Produce collection of plots used to visalise estimated shear bias after correction

    """
    
    visualise(currents, options, plot_all=False)
    
    bias_along_glider, bias_across_glider = regress_bias(currents, options)
    
    currents = correct_bias(currents, options, bias_along_glider, bias_across_glider)
    
    visualise(currents, options)
    
    return currents